import unittest
from unittest.mock import MagicMock, patch
import sys

from lifeguard.repositories import (
    IMPLEMENTATIONS,
    HistoryRepository,
    NotificationRepository,
    ValidationRepository,
    declare_implementation,
)

sys.path.append("tests/fixtures")

NAMES = ["history", "notification", "validation"]


class TestNotificationRepository(unittest.TestCase):
    def setUp(self):
        for name in NAMES:
            if name in IMPLEMENTATIONS:
                IMPLEMENTATIONS.pop(name)

        self.implementation = MagicMock(name="implementation")

        with patch("lifeguard.repositories.logger"):
            implementation_class = MagicMock(name="implementation_class")
            implementation_class.__name__ = "mocked_implementation"
            implementation_class.return_value = self.implementation

            declare_implementation("notification", implementation_class)

        self.notification_repository = NotificationRepository()

    def test_save_last_notification_for_a_validation(self):
        result = MagicMock(name="result")
        self.notification_repository.save_last_notification_for_a_validation(result)
        self.implementation.save_last_notification_for_a_validation.assert_called_with(
            result
        )

    def test_validation_repository_fetch_last_validation_result(self):
        validation_name = MagicMock(name="validation_name")
        self.notification_repository.fetch_last_notification_for_a_validation(
            validation_name
        )
        self.implementation.fetch_last_notification_for_a_validation.assert_called_with(
            validation_name
        )


class TestHistoryRepository(unittest.TestCase):
    def setUp(self):
        for name in NAMES:
            if name in IMPLEMENTATIONS:
                IMPLEMENTATIONS.pop(name)

        self.implementation = MagicMock(name="implementation")

        with patch("lifeguard.repositories.logger"):
            implementation_class = MagicMock(name="implementation_class")
            implementation_class.__name__ = "mocked_implementation"
            implementation_class.return_value = self.implementation

            declare_implementation("history", implementation_class)

        self.history_repository = HistoryRepository()

    def test_append_notification(self):
        notification_occurrence = MagicMock(name="notification_occurrence")
        self.history_repository.append_notification(notification_occurrence)
        self.implementation.append_notification.assert_called_with(
            notification_occurrence
        )

    def test_count_notifications(self):
        result = self.history_repository.count_notifications()

        self.implementation.count_notifications.assert_called_with(None, None, {})
        self.assertIsNotNone(result)

    def test_fetch_notifications(self):
        start_interval = MagicMock(name="start_interval")
        end_interval = MagicMock(name="end_interval")
        filters = MagicMock(name="filters")

        result = self.history_repository.fetch_notifications(
            start_interval, end_interval, filters
        )

        self.implementation.fetch_notifications.assert_called_with(
            start_interval, end_interval, filters, None, None
        )
        self.assertIsNotNone(result)


class TestValidationRepositories(unittest.TestCase):
    def setUp(self):
        for name in NAMES:
            if name in IMPLEMENTATIONS:
                IMPLEMENTATIONS.pop(name)

        self.implementation = MagicMock(name="implementation")
        self.implementation.__name__ = "mocked_implementation"

        with patch("lifeguard.repositories.logger"):
            implementation_class = MagicMock(name="implementation_class")
            implementation_class.__name__ = "mocked_implementation"
            implementation_class.return_value = self.implementation

            declare_implementation("validation", implementation_class)

        self.validation_repository = ValidationRepository()

    def test_validation_repository_save_validation_result(self):
        result = MagicMock(name="result")
        self.validation_repository.save_validation_result(result)
        self.implementation.save_validation_result.assert_called_with(result)

    def test_validation_repository_fetch_last_validation_result(self):
        validation_name = MagicMock(name="validation_name")
        self.validation_repository.fetch_last_validation_result(validation_name)
        self.implementation.fetch_last_validation_result.assert_called_with(
            validation_name
        )

    def test_validation_repository_fetch_all_validation_results(self):
        self.validation_repository.fetch_all_validation_results()
        self.implementation.fetch_all_validation_results.assert_called()

    def test_validation_repository_delete_validation_result(self):
        validation_name = MagicMock(name="validation_name")
        self.validation_repository.delete_validation_result(validation_name)
        self.implementation.delete_validation_result(validation_name)


class TestRepositoriesFunctions(unittest.TestCase):
    def setUp(self):
        if "test" in IMPLEMENTATIONS:
            IMPLEMENTATIONS.pop("test")
        self.implementation = MagicMock(name="implementation")
        self.implementation.__name__ = "test"

    @patch("lifeguard.repositories.logger")
    def test_warning_when_an_implementation_is_overwrited(self, mock_logger):
        declare_implementation("test", self.implementation)
        declare_implementation("test", self.implementation)
        mock_logger.warning.assert_called_with(
            "overwriting implementation for respository %s", "test"
        )

    @patch("lifeguard.repositories.logger")
    def test_info_when_an_implementation_is_declared(self, mock_logger):
        declare_implementation("test", self.implementation)
        mock_logger.info.assert_called_with(
            "loading implementation %s for repository %s",
            "test",
            "test",
        )



import json
import unittest
from unittest.mock import MagicMock, patch

from lifeguard import NORMAL, PROBLEM
from lifeguard.server import (
    execute_validation,
    validation_endpoint,
    get_status,
    get_status_complete,
)
from lifeguard.validations import ValidationResponse

test_validation = MagicMock(name="test_validation")
test_validation.return_value = ValidationResponse(NORMAL, {})

VALIDATIONS = {"test_validation": {"ref": test_validation}}


class TestServer(unittest.TestCase):
    @patch("lifeguard.server.make_response")
    @patch("lifeguard.server.VALIDATIONS", VALIDATIONS)
    def test_execute_validation(self, mock_make_response):
        mock_response = MagicMock(name="mock_response")
        mock_make_response.return_value = mock_response

        response = execute_validation("test_validation")
        self.assertEqual(response, mock_response)

    @patch("lifeguard.server.make_response")
    @patch("lifeguard.server.traceback")
    @patch("lifeguard.server.logger")
    @patch("lifeguard.server.VALIDATIONS", VALIDATIONS)
    def test_execute_validation_with_error(
        self, mock_logger, mock_traceback, mock_make_response
    ):
        mock_traceback.format_exc.return_value = "traceback"
        mock_make_response.side_effect = Exception("exception")

        response = execute_validation("test_validation")
        self.assertEqual(response, json.dumps({"error": "traceback"}))
        mock_logger.error.assert_called_with(
            "error on execute validation %s",
            "test_validation",
            extra={"traceback": "traceback"},
        )

    @patch("lifeguard.server.make_response")
    @patch("lifeguard.server.ValidationRepository")
    @patch("lifeguard.server.request")
    def test_validation_endpoint_get_result(
        self, mock_request, mock_validation_repository, mock_make_response
    ):
        mock_request.method = "GET"
        mock_response = MagicMock(name="mock_response")
        mock_make_response.return_value = mock_response

        mock_repository_instance = MagicMock(name="mock_repository_instance")
        mock_validation_repository.return_value = mock_repository_instance
        mock_repository_instance.fetch_last_validation_result.return_value = (
            ValidationResponse(NORMAL, {})
        )

        response = validation_endpoint("test_validation")
        self.assertEqual(response, mock_response)

    @patch("lifeguard.server.make_response")
    @patch("lifeguard.server.ValidationRepository")
    @patch("lifeguard.server.request")
    def test_validation_endpoint_delete_result(
        self, mock_request, mock_validation_repository, mock_make_response
    ):
        mock_request.method = "DELETE"

        mock_repository_instance = MagicMock(name="mock_repository_instance")
        mock_validation_repository.return_value = mock_repository_instance

        validation_endpoint("test_validation")
        mock_make_response.assert_called_with('{"status": "ok"}')
        mock_repository_instance.delete_validation_result.assert_called_with(
            "test_validation"
        )

    @patch("lifeguard.server.make_response")
    @patch("lifeguard.server.ValidationRepository")
    def test_get_status(self, mock_validation_repository, mock_make_response):
        mock_response = MagicMock(name="mock_response")
        mock_make_response.return_value = mock_response

        mock_repository_instance = MagicMock(name="mock_repository_instance")
        mock_validation_repository.return_value = mock_repository_instance
        mock_repository_instance.fetch_all_validation_results.return_value = [
            ValidationResponse(PROBLEM, {}, validation_name="test_validation")
        ]

        response = get_status()
        mock_make_response.assert_called_with(
            '{"status": "PROBLEM", "validations": [{"validation_name": "test_validation", "status": "PROBLEM", "details": {}, "settings": null}]}'
        )
        self.assertEqual(response, mock_response)

    @patch("lifeguard.server.make_response")
    @patch("lifeguard.server.ValidationRepository")
    def test_get_status_validations_empty_when_normal(
        self, mock_validation_repository, mock_make_response
    ):
        mock_response = MagicMock(name="mock_response")
        mock_make_response.return_value = mock_response

        mock_repository_instance = MagicMock(name="mock_repository_instance")
        mock_validation_repository.return_value = mock_repository_instance
        mock_repository_instance.fetch_all_validation_results.return_value = [
            ValidationResponse(NORMAL, {})
        ]

        response = get_status()
        mock_make_response.assert_called_with('{"status": "NORMAL", "validations": []}')
        self.assertEqual(response, mock_response)

    @patch("lifeguard.server.make_response")
    @patch("lifeguard.server.ValidationRepository")
    def test_get_status_complete(self, mock_validation_repository, mock_make_response):
        mock_response = MagicMock(name="mock_response")
        mock_make_response.return_value = mock_response

        mock_repository_instance = MagicMock(name="mock_repository_instance")
        mock_validation_repository.return_value = mock_repository_instance
        mock_repository_instance.fetch_all_validation_results.return_value = [
            ValidationResponse(NORMAL, {}, validation_name="test_validation")
        ]

        response = get_status_complete()
        mock_make_response.assert_called_with(
            '{"status": "NORMAL", "validations": [{"validation_name": "test_validation", "status": "NORMAL", "details": {}, "settings": null}]}'
        )
        self.assertEqual(response, mock_response)

    @patch("lifeguard.server.make_response")
    @patch("lifeguard.server.ValidationRepository")
    @patch("lifeguard.server.traceback")
    def test_get_status_on_exception(
        self, mock_traceback, mock_validation_repository, mock_make_response
    ):
        mock_traceback.format_exc.return_value = "traceback"

        mock_response = MagicMock(name="mock_response")
        mock_make_response.return_value = mock_response

        mock_validation_repository.side_effect = [Exception("error")]

        response = get_status_complete()
        self.assertEqual(response, mock_response)



import unittest

from unittest.mock import patch, call

from lifeguard.settings import (
    PERMANENT_SESSION_LIFETIME,
    LIFEGUARD_SERVER_PORT,
    LIFEGUARD_DIRECTORY,
    LIFEGUARD_RUN_ONLY_VALIDATIONS,
    LIFEGUARD_SKIP_VALIDATIONS,
    LOG_LEVEL,
    SETTINGS_MANAGER,
    SettingsManager,
    AttributeNotFoundInSettings,
)


class SettingsTest(unittest.TestCase):
    def test_server_port(self):
        self.assertEqual(LIFEGUARD_SERVER_PORT, 5567)
        self.assertEqual(
            SETTINGS_MANAGER.settings["LIFEGUARD_SERVER_PORT"]["description"],
            "Lifeguard server port number",
        )

    def test_log_level(self):
        self.assertEqual(LOG_LEVEL, "INFO")
        self.assertEqual(
            SETTINGS_MANAGER.settings["LIFEGUARD_LOG_LEVEL"]["description"],
            "Sets the Lifeguard's core log level",
        )

    def test_lifeguard_directory(self):
        self.assertEqual(LIFEGUARD_DIRECTORY, ".")
        self.assertEqual(
            SETTINGS_MANAGER.settings["LIFEGUARD_DIRECTORY"]["description"],
            "Location of validations and others resources",
        )

    def test_get_a_string_list(self):
        settings = SettingsManager(
            {
                "LIST": {
                    "default": "item1,item2,",
                    "type": "list",
                    "description": "a str list",
                }
            }
        )
        self.assertEqual(["item1", "item2"], settings.read_value("LIST"))

    def test_lifeguard_run_only_validations_default(self):
        self.assertEqual(LIFEGUARD_RUN_ONLY_VALIDATIONS, [])
        self.assertEqual(
            SETTINGS_MANAGER.settings["LIFEGUARD_RUN_ONLY_VALIDATIONS"]["description"],
            "A comma separated list with validations name to be run",
        )

    def test_get_allowed_validations_returns_1_validation(self):
        settings = SettingsManager(
            {
                "LIFEGUARD_RUN_ONLY_VALIDATIONS": {
                    "default": "validation1",
                    "type": "validation_list",
                    "description": "allowed validations",
                }
            }
        )
        self.assertEqual(
            ["validation1"],
            settings.read_value("LIFEGUARD_RUN_ONLY_VALIDATIONS"),
        )

    def test_get_allowed_validations_returns_2_validations(self):
        settings = SettingsManager(
            {
                "LIFEGUARD_RUN_ONLY_VALIDATIONS": {
                    "default": "validation1,validation2",
                    "type": "validation_list",
                    "description": "allowed validations",
                }
            }
        )
        self.assertEqual(
            ["validation1", "validation2"],
            settings.read_value("LIFEGUARD_RUN_ONLY_VALIDATIONS"),
        )

    def test_lifeguard_skip_validations_default(self):
        self.assertEqual(LIFEGUARD_SKIP_VALIDATIONS, [])
        self.assertEqual(
            SETTINGS_MANAGER.settings["LIFEGUARD_SKIP_VALIDATIONS"]["description"],
            "A comma separated list with validations name to be skipped",
        )

    def test_get_default_value_for_dynamic_attribute(self):
        settings = SettingsManager(
            {
                r"DYNAMIC_\w+_ATTRIBUTE": {
                    "default": "default_value",
                    "description": "dynammic descritpion",
                }
            }
        )

        self.assertEqual(settings.read_value("DYNAMIC_TEST_ATTRIBUTE"), "default_value")

    def test_setting_with_declared_type_returns_converted_value(self):
        settings = SettingsManager(
            {
                "LIFEGUARD_SERVER_PORT": {
                    "default": "5567",
                    "type": "int",
                    "description": "Lifeguard server port number",
                }
            }
        )
        current_value = settings.read_value("LIFEGUARD_SERVER_PORT")
        self.assertEqual(current_value, 5567)

    def test_setting_with_invalid_declared_type_returns_original_value(self):
        settings = SettingsManager(
            {
                "LIFEGUARD_SERVER_PORT": {
                    "default": "5567",
                    "type": True,
                    "description": "Lifeguard server port number",
                }
            }
        )
        current_value = settings.read_value("LIFEGUARD_SERVER_PORT")
        self.assertEqual(current_value, "5567")

    def test_get_default_value_with_declared_type_for_dynamic_attribute(self):
        settings = SettingsManager(
            {
                "BOOL_SETTING": {
                    "default": "false",
                    "type": "bool",
                    "description": "dynammic descritpion",
                }
            }
        )
        self.assertEqual(False, settings.read_value("BOOL_SETTING"))

    @patch("lifeguard.settings.environ")
    def test_get_value_for_dynamic_attribute_from_env(self, mock_environ):
        mock_environ.get.return_value = "from env"
        settings = SettingsManager(
            {
                r"DYNAMIC_\w+_ATTRIBUTE": {
                    "default": "default_value",
                    "description": "dynammic descritpion",
                }
            }
        )
        self.assertEqual(settings.read_value("DYNAMIC_TEST_ATTRIBUTE"), "from env")
        mock_environ.get.assert_has_calls(
            [
                call("DYNAMIC_\\w+_ATTRIBUTE", "default_value"),
                call("DYNAMIC_TEST_ATTRIBUTE", "default_value"),
            ]
        )

    @patch("lifeguard.settings.environ")
    def test_get_value_with_type_for_dynamic_attribute_from_env(self, mock_environ):
        mock_environ.get.return_value = "false"
        settings = SettingsManager(
            {
                "DYNAMIC_TEST_ATTRIBUTE": {
                    "default": "default_value",
                    "type": "bool",
                    "description": "dynammic descritpion",
                }
            }
        )

        self.assertEqual(False, settings.read_value("DYNAMIC_TEST_ATTRIBUTE"))
        mock_environ.get.assert_has_calls(
            [
                call("DYNAMIC_TEST_ATTRIBUTE", "default_value"),
            ]
        )

    def test_raise_exception_if_pattern_not_found(self):
        settings = SettingsManager(
            {
                r"DYNAMIC_\w+_ATTRIBUTE": {
                    "default": "default_value",
                    "description": "dynammic descritpion",
                }
            }
        )
        with self.assertRaises(AttributeNotFoundInSettings):
            settings.read_value("DYNAMIC__ATTRIBUTE")

    def test_permanent_session_lifetime(self):
        self.assertEqual(PERMANENT_SESSION_LIFETIME, 2678400)
        self.assertEqual(
            SETTINGS_MANAGER.settings["PERMANENT_SESSION_LIFETIME"]["type"],
            "int",
        )
        self.assertEqual(
            SETTINGS_MANAGER.settings["PERMANENT_SESSION_LIFETIME"]["description"],
            "Session expiration time",
        )



"""
Settings for application logging
"""
import json
import logging
import sys
import time
from datetime import datetime
from logging.handlers import RotatingFileHandler

from lifeguard.settings import LOG_LEVEL, LOG_FILE, LOG_BACKUP_COUNT, LOG_MAX_BYTES


class CustomFormatter(logging.Formatter):
    def __init__(self):
        super(CustomFormatter, self).__init__()

    def format(self, record):
        message_object = {
            "level": record.levelname,
            "message": {
                "fileName": record.filename,
                "lineNumber": record.lineno,
                "appName": "lifeguard",
                "message": record.msg % record.args,
            },
            "metadata": {
                "timestamp": datetime.now().strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z",
                "processName": sys.argv[0],
            },
            "time": int(time.time()),
        }

        if record.traceback:
            message_object["metadata"]["traceback"] = record.traceback

        return json.dumps(message_object)


class CustomLogger(logging.Logger):
    """
    Custom logger implementation
    """

    def _log(self, level, msg, args, exc_info=None, extra=None, stack_info=None):
        if extra is None:
            extra = {}
        if "traceback" not in extra:
            extra["traceback"] = ""

        super(CustomLogger, self)._log(level, msg, args, exc_info, extra, stack_info)


logging.setLoggerClass(CustomLogger)
lifeguard_logger = logging.getLogger("lifeguard")
lifeguard_logger.propagate = 0
lifeguard_logger.setLevel(getattr(logging, LOG_LEVEL))

if LOG_FILE:
    LIFEGUARD_FILE_HANDLER = RotatingFileHandler(
        LOG_FILE, maxBytes=LOG_MAX_BYTES, backupCount=LOG_BACKUP_COUNT
    )
    LIFEGUARD_FILE_HANDLER.setFormatter(CustomFormatter())
    lifeguard_logger.addHandler(LIFEGUARD_FILE_HANDLER)


LIFEGUARD_HANDLER = logging.StreamHandler()
LIFEGUARD_HANDLER.setFormatter(CustomFormatter())

lifeguard_logger.addHandler(LIFEGUARD_HANDLER)



"""
Wrapper for http requests
"""
import requests

from lifeguard.settings import HTTP_PROXY, HTTPS_PROXY


def get(url, headers=None, auth=None):
    """
    :param url:
    :param headers:

    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """
    request_args = {"url": url, "headers": headers, "auth": auth}
    __append_proxies(request_args)
    return requests.get(**request_args)


def post(url, data=None, headers=None, auth=None):
    """
    :param url:
    :param data:
    :param headers:

    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    request_args = {"url": url, "headers": headers, "data": data, "auth": auth}
    __append_proxies(request_args)
    return requests.post(**request_args)


def __append_proxies(request_args):
    proxies = {
        "http": HTTP_PROXY,
        "https": HTTPS_PROXY,
    }

    keys_to_remove = [key for key in proxies.keys() if not proxies[key]]
    for key in keys_to_remove:
        proxies.pop(key)

    if proxies:
        request_args["proxies"] = proxies



import unittest
from unittest.mock import patch, MagicMock

from lifeguard import NORMAL, PROBLEM
from lifeguard.actions.email import send_email, EMAIL_NOTIFICATIONS


class ActionEmailTest(unittest.TestCase):
    @patch("lifeguard.actions.email.logger")
    @patch("lifeguard.actions.email.smtplib")
    def test_send_email(self, mock_smtplib, mock_logger):
        mock_session = MagicMock(name="session")
        mock_smtplib.SMTP.return_value = mock_session

        validation_response = MagicMock(name="validation_response")
        validation_response.validation_name = "validation_with_problem"
        validation_response.status = PROBLEM
        validation_response.details = {"status": PROBLEM}
        send_email(
            validation_response,
            {
                "email": {
                    "subject": "subject example",
                    "receivers": [{"name": "name", "email": "email@server.com"}],
                    "send_in": [PROBLEM],
                }
            },
        )

        mock_smtplib.SMTP.assert_called_with("smtp_server", port="smtp_port")
        mock_session.starttls.assert_called()
        mock_session.login.assert_called_with("smtp_user", "smtp_passwd")
        mock_session.sendmail.assert_called_with(
            "smtp_user",
            ["email@server.com"],
            "From: smtp_user\nTo: name <email@server.com>\nMIME-Version: 1.0\nContent-type: text/html\nSubject: subject example\n\n<h1>Validation details</h1>\n\n<quote>\n{'status': 'PROBLEM'}\n</quote>\n\n<hr/>\n<p>\n    See more details in Lifeguard Dashboard: http://localhost:5567.\n</p>\n",
        )
        mock_session.quit.assert_called()
        mock_logger.info.assert_called_with(
            "sending email %s to %s",
            "From: smtp_user\nTo: name <email@server.com>\nMIME-Version: 1.0\nContent-type: text/html\nSubject: subject example\n\n<h1>Validation details</h1>\n\n<quote>\n{'status': 'PROBLEM'}\n</quote>\n\n<hr/>\n<p>\n    See more details in Lifeguard Dashboard: http://localhost:5567.\n</p>\n",
            ["email@server.com"],
        )
        self.assertTrue(validation_response.validation_name in EMAIL_NOTIFICATIONS)

    @patch("lifeguard.actions.email.logger")
    @patch("lifeguard.actions.email.smtplib")
    def test_not_send_email(self, mock_smtplib, mock_logger):
        mock_session = MagicMock(name="session")
        mock_smtplib.SMTP.return_value = mock_session

        validation_response = MagicMock(name="validation_response")
        validation_response.validation_name = "validation_test"
        validation_response.status = NORMAL
        validation_response.details = {}

        send_email(
            validation_response,
            {
                "email": {
                    "subject": "subject example",
                    "receivers": [{"name": "name", "email": "email@server.com"}],
                }
            },
        )

        mock_logger.info.assert_not_called()

    @patch("lifeguard.actions.email.logger")
    @patch("lifeguard.actions.email.smtplib")
    def test_remove_validation_from_sent_list(self, mock_smtplib, mock_logger):
        mock_session = MagicMock(name="session")
        mock_smtplib.SMTP.return_value = mock_session

        validation_response = MagicMock(name="validation_response")
        validation_response.validation_name = "validation_in_list"
        validation_response.status = NORMAL
        validation_response.details = {}

        EMAIL_NOTIFICATIONS["validation_in_list"] = {}

        send_email(
            validation_response,
            {
                "email": {
                    "subject": "subject example",
                    "receivers": [{"name": "name", "email": "email@server.com"}],
                }
            },
        )

        mock_logger.info.assert_not_called()
        self.assertTrue("validation_in_list" not in EMAIL_NOTIFICATIONS)

    @patch("lifeguard.actions.email.traceback")
    @patch("lifeguard.actions.email.logger")
    @patch("lifeguard.actions.email.smtplib")
    def test_error_on_send_email(self, mock_smtplib, mock_logger, mock_traceback):
        mock_session = MagicMock(name="session")
        mock_smtplib.SMTP.return_value = mock_session

        mock_session.starttls.side_effect = Exception("tls error")

        validation_response = MagicMock(name="validation_response")
        validation_response.status = PROBLEM
        validation_response.details = {}

        mock_traceback.format_exc.return_value = "format_exc"

        send_email(
            validation_response,
            {
                "email": {
                    "subject": "subject example",
                    "receivers": [{"name": "name", "email": "email@server.com"}],
                    "send_in": [PROBLEM],
                }
            },
        )

        mock_logger.error.assert_called_with(
            "error on send email %s to %s",
            "From: smtp_user\nTo: name <email@server.com>\nMIME-Version: 1.0\nContent-type: text/html\nSubject: subject example\n\n<h1>Validation details</h1>\n\n<quote>\n{}\n</quote>\n\n<hr/>\n<p>\n    See more details in Lifeguard Dashboard: http://localhost:5567.\n</p>\n",
            ["email@server.com"],
            extra={"traceback": "format_exc"},
        )



"""
Base of action used to notification
"""
from copy import deepcopy
from datetime import datetime
import jinja2
import json

from lifeguard.statuses import NORMAL, PROBLEM
from lifeguard.logger import lifeguard_logger as logger
from lifeguard.notifications import (
    NOTIFICATION_METHODS,
    INIT_THREAD_MESSAGE_NOTIFICATION,
    CLOSE_THREAD_MESSAGE_NOTIFICATION,
    SINGLE_MESSAGE_NOTIFICATION,
    UPDATE_THREAD_MESSAGE_NOTIFICATION,
    NotificationOccurrence,
    NotificationStatus,
)
from lifeguard.repositories import HistoryRepository, NotificationRepository
from lifeguard.settings import LIFEGUARD_APPEND_NOTIFICATION_TO_HISTORY


def __merge_details_and_data(validation_response, notification_data):
    data = deepcopy(validation_response.details)
    data.update(notification_data)
    return data


def __get_content(validation_response, settings):
    template_origin = (
        "template"
        if not (validation_response.details or {}).get("use_error_template", False)
        else "template_error"
    )

    if template_origin in settings.get("notification", {}):
        template = jinja2.Template(settings["notification"][template_origin])
        notification_data = (
            (validation_response.settings or {}).get("notification", {}).get("data", {})
        )
        if isinstance(notification_data, list):
            return [
                template.render(**__merge_details_and_data(validation_response, data))
                for data in notification_data
            ]
        return template.render(
            **__merge_details_and_data(
                validation_response,
                notification_data,
            )
        )
    return json.dumps(validation_response.details)


def __get_notification_methods(settings):
    disabled = settings.get("notification", {}).get("disabled", [])
    return [method for method in NOTIFICATION_METHODS if method.name not in disabled]


def notify_in_thread(validation_response, settings):
    """
    Create a new thread to notify problem status.
    When status returns to NORMAL thread should be closed.

    :param validation_response: a validation response
    :param settings: validation settings
    """
    repository = NotificationRepository()
    last_notification_status = repository.fetch_last_notification_for_a_validation(
        validation_response.validation_name
    )

    if not last_notification_status and validation_response.status != PROBLEM:
        return

    content = __get_content(validation_response, settings)

    if not last_notification_status:
        thread_ids = {}
        for notification_method in __get_notification_methods(settings):
            thread_ids[notification_method.name] = notification_method.init_thread(
                deepcopy(content), settings
            )

        last_notification_status = NotificationStatus(
            validation_response.validation_name, thread_ids
        )
        __append_notification(
            validation_response,
            settings,
            INIT_THREAD_MESSAGE_NOTIFICATION,
        )
    else:
        for notification_method in __get_notification_methods(settings):
            if notification_method.name in last_notification_status.thread_ids:
                __send_notification_in_thread(
                    last_notification_status,
                    notification_method,
                    deepcopy(content),
                    settings,
                    validation_response,
                )
    repository.save_last_notification_for_a_validation(last_notification_status)


def __send_notification_in_thread(
    last_notification_status,
    notification_method,
    content,
    settings,
    validation_response,
):
    thread_id = last_notification_status.thread_ids[notification_method.name]
    if validation_response.status == NORMAL:
        notification_method.close_thread(thread_id, content, settings)
        last_notification_status.close()
        __append_notification(
            validation_response,
            settings,
            CLOSE_THREAD_MESSAGE_NOTIFICATION,
        )
    else:
        last_notification_was_in_seconds = (
            datetime.now() - last_notification_status.last_notification
        ).seconds
        interval = settings.get("notification", {}).get("update_thread_interval", 0)

        if last_notification_was_in_seconds >= interval:
            logger.debug("updating notification")
            notification_method.update_thread(thread_id, content, settings)
            last_notification_status.update()
            __append_notification(
                validation_response,
                settings,
                UPDATE_THREAD_MESSAGE_NOTIFICATION,
            )


def __append_notification(validation_response, settings, notification_type):
    notification_settings = settings.get("notification", {})
    add_to_history = notification_settings.get(
        "add_to_history", LIFEGUARD_APPEND_NOTIFICATION_TO_HISTORY
    )
    if add_to_history:
        HistoryRepository().append_notification(
            NotificationOccurrence(
                validation_name=validation_response.validation_name,
                details=validation_response.details,
                status=validation_response.status,
                notification_type=notification_type,
            )
        )


def notify_in_single_message(validation_response, settings):
    """
    This action sends single message on a new thread.

    To send notification validation response should have:
    {
        "notification": {"notify": True}
    }

    :param validation_response: a validation response
    :param settings: validation settings
    """

    response_settings = validation_response.settings or {}
    notification_settings = response_settings.get("notification", {})
    should_notify = notification_settings.get("notify", False)

    if should_notify:
        content = __get_content(validation_response, settings)
        for notification_method in __get_notification_methods(settings):
            notification_method.send_single_message(deepcopy(content), settings)

        __append_notification(
            validation_response, settings, SINGLE_MESSAGE_NOTIFICATION
        )
